#This program generate a Non Local Means speckle filter
#by means of Shannon Entropy

# Setup and libraries
library(maxLik)
library(POT)
library(OpenImageR)
library(corpcor)
require(svMisc)
set.seed(1234567890, kind = "Mersenne-Twister")

#Shannon entropy for GI0(alpha,gamma)
CalculaEntropiaGI0Sh<-function(alpha, gama)
{entrop <- (alpha-1)/alpha-log(-alpha/gama)
return(entrop)
}
#Renyi entropy for GI0
CalculaEntropiaGI0Reny<-function(alpha, gama, beta)
{# beta between 0 y 1
  entrop <- beta/(1-beta)*log(-alpha/gama) + 1/(1-beta)*log(-gama/(beta*(alpha-1)+1))
  return(entrop)
}

#Fisher Information matrix
MatrizInformacionFisher <- function(alpha, gama)
{
  a11 <- 1/alpha^2
  a12 <- 1/(gama*(1-alpha))
  a22 <- alpha/((alpha - 2)*gama^2)
  K <- matrix(c(a11,a12,a12,a22), ncol = 2)
  return(K)
}

deltaShannon <- function(alpha, gama)
{
  d1 <- 1/alpha^2 - 1/alpha
  d2 <- 1/gama
  return (c(d1,d2))
}
#Variance
VarianzaEntropiaSh <- function(alpha,gama)
{
  K <-MatrizInformacionFisher(alpha, gama)
  delta <- deltaShannon(alpha, gama)
  if(det(K) > 0.000000001){
    V <- t(delta) %*% solve(K)%*% delta}
  else{
    V = pseudoinverse(K)}
  
  return(V)
}

deltaRenyi <- function(alpha, gama, beta)
{
  d1 <- (beta/(1-beta))*( 1/alpha -1/(beta*(alpha-1) + 1))
  d2 <- 1/gama
  return (c(d1,d2))
}

#Variance of Renyi
VarianzaEntropiaRenyi <- function(alpha,gama,beta)
{
  K <-MatrizInformacionFisher(alpha, gama)
  delta <- deltaRenyi(alpha, gama,beta)
  if(det(K) > 0.000000001){
    V <- t(delta) %*% solve(K)%*% delta}
  else{
    V = pseudoinverse(K)}
  
  return(V)
}

#Smooth function
SmootherStep <- function(x) {
  ifelse((x <= 0), 0.2, 
         ifelse((x > 0 & x <= 1),  6*x^5 - 15*x^4 + 10*x^3,
                ifelse((x > 1), 1, NA)))
}

#p0 = eta
PtoW <- function(p, p0, k) {
  SmootherStep((p-p0/k)/(p0-p0/k))
}

#Similarity measure
#tam: size of the window
#Img: original Image
#eta: significance level
MedidaDeSimilaridad <-function(tam, Img,eta)
{
  d = dim(Img)
  fin = floor(d[1]/tam)
  division.fila = seq(1, d[1] + 1) 
  division.col = seq(1, d[2] + 1)  
  lr = length(division.fila)
  lk = length(division.col)
  estimates <-rep(0,(lr-1)*(lk-1)*2) 
  estimates<-matrix(estimates,ncol= 2)
  
  estimatesPOT <-rep(0,(lr-1)*(lk-1)*2) 
  estimatesPOT<-matrix(estimatesPOT,ncol= 2) 
  
  DifEntropiasSh <-rep(0,(lr-1)*(lk-1)) 
  VarianzaEntropiaSh <-rep(0,(lr-1)*(lk-1)) 
  

  VentCentral <- Img[((floor(d[1]/2)+1)-(floor(tam/2))): ((floor(d[1]/2)+1)+(floor(tam/2))),
                     ((floor(d[2]/2)+1)-(floor(tam/2))): ((floor(d[2]/2)+1)+(floor(tam/2)))]
  
  #Parameter estimation with POT
  shape.mle = fitgpd(VentCentral, threshold=0, "mle")$param[[2]] 
  
  scale.mle = fitgpd(VentCentral, threshold=0, "mle")$param[[1]] 
  estCentroImagenPOT <-rep(0,2)
  estCentroImagenPOT[1] <- min(-1.1,max(-15,-1/shape.mle))
  estCentroImagenPOT[2] <- max(0.1,scale.mle/shape.mle)
  
  EntropiaSh <- rep(0,(lr-1)*(lk-1))
  
  entropiaCentroImagenSh <- CalculaEntropiaGI0Sh(estCentroImagenPOT[1], estCentroImagenPOT[2]) #entropia para el centro de la imagen
  VarianzaShCentral <- VarianzaEntropiaSh(estCentroImagenPOT[1], estCentroImagenPOT[2])
  
  S <- rep(0,(lr-1)*(lk-1)) 
  pv <- rep(0,(lr-1)*(lk-1))
  result<-rep(0,(lr-1)*(lk-1))
  result<-matrix(result, ncol = d[2])
  beta = 40*eta

  for(r in (1+floor(tam/2)):((lr-1)-floor(tam/2)))
  {
    for(k in (1+floor(tam/2)):((lk-1)-floor(tam/2)))
    {
      newim <- Img[(r-floor(tam/2)):(r+floor(tam/2)),(k-floor(tam/2)):(k+floor(tam/2))] #toma una ventanita
      
      shape.mle <- fitgpd(newim, threshold=0, "mle")$param[[2]]
      scale.mle <- fitgpd(newim, threshold=0, "mle")$param[[1]] 
      
      estimatesPOT[(r-floor(tam/2)-1)*(lr-1)+k,1] <- min(-1.1, max(-15,-1/shape.mle))
      estimatesPOT[(r-floor(tam/2)-1)*(lr-1)+k,2] <- max(0.1, scale.mle/shape.mle)
      
      
      #Calcula La entropia de la GI0 para los cuadraditos
      EntropiaSh[(r-floor(tam/2)-1)*(lr-1)+k] = CalculaEntropiaGI0Sh(estimatesPOT[(r-floor(tam/2)-1)*(lr-1)+k,1],
                                                                     estimatesPOT[(r-floor(tam/2)-1)*(lr-1)+k,2])
      
      VarianzaEntropiaSh[(r-floor(tam/2)-1)*(lr-1)+k] <- VarianzaEntropiaSh(estimatesPOT[(r-floor(tam/2)-1)*(lr-1)+k,1],
                                                                            estimatesPOT[(r-floor(tam/2)-1)*(lr-1)+k,2])

      S[(r-floor(tam/2)-1)*(lr-1)+k] <- tam^2*(EntropiaSh[(r-floor(tam/2)-1)*(lr-1)+k] -  entropiaCentroImagenSh)^2/
        (VarianzaEntropiaSh[(r-floor(tam/2)-1)*(lr-1)+k] + VarianzaShCentral)
  
    
      pv[(r-floor(tam/2)-1)*(lr-1)+k] <- pchisq(S[(r-floor(tam/2)-1)*(lr-1)+k],1)
      
      result[r,k] <- PtoW(pv[(r-floor(tam/2)-1)*(lr-1)+k], eta, 2)
    }
  }

  result = result[(floor(tam/2)+1):(d[1]-floor(tam/2)), (floor(tam/2)+1):(d[1]-floor(tam/2))]
  norma<- sum(apply(result,1,sum))
  result = result/(norma + 0.000001)
  Mult <- result*Img[(1+floor(tam/2)):(d[1]-floor(tam/2)), (1+floor(tam/2)):(d[2]-floor(tam/2))]
  Conv <- sum(Mult)
  
  return(Conv)
}



#Main Program
#tamVG: Size of the large window
#tam: size of the small window
#Img: original image
#eta: significance level
#Example:
#FiltroConElpvalor(15, 9, Img, 0.2)
FiltroConElpvalor<- function(tamVG, tam, Img, eta)
{
  start_time <- Sys.time()
  
  srk = dim(Img)
  ImgFilt = Img
  
  for(r in  (1+floor(tamVG/2)):(srk[1]-floor(tamVG/2)))
  {
    progress(r, max.value = (srk[1]-floor(tamVG/2)))
    
    for(k in (1+floor(tamVG/2)):(srk[2]-floor(tamVG/2)))
    {
      ImgLocal <- Img[(r-floor(tamVG/2)):(r+floor(tamVG/2)),(k-floor(tamVG/2)):(k+floor(tamVG/2))] #toma la ventana grande
      PixelNuevo <- MedidaDeSimilaridad(tam, ImgLocal, eta)
      ImgFilt[r,k] <- PixelNuevo
    }
  }
  end_time <- Sys.time()
  
  cat("Elapsed time:", end_time-start_time, "seconds")
  return(ImgFilt)
}



